"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateCompileBasedDecoders = void 0;
const fs_1 = require("fs");
const path_1 = __importDefault(require("path"));
const prettier_1 = require("prettier");
const generation_utils_1 = require("./generation-utils");
const decodersFileTemplate = `
/* eslint-disable */
import Ajv from 'ajv';
$Imports
import { Decoder } from './helpers';
import { validateJson } from './validate';
import { $ModelImports } from './models';
import jsonSchema from './schema.json';
const ajv = new Ajv({ strict: false });
ajv.compile(jsonSchema);
$Formats
// Decoders
$Decoders
`;
const decoderTemplate = `
export const $DecoderName: Decoder<$Class> = {
  definitionName: '$Class',
  schemaRef: '#/definitions/$Class',
  decode(json: unknown): $Class {
    const schema = ajv.getSchema($DecoderName.schemaRef);
    if (!schema) {
      throw new Error(\`Schema \${$DecoderName.definitionName} not found\`);
    }
    return validateJson(json, schema, $DecoderName.definitionName);
  }
}
`;
function generateCompileBasedDecoders(definitionNames, addFormats, formatOptions, outDirs, prettierOptions) {
    const decoders = definitionNames
        .map(definitionName => decoderTemplate
        .replace(/\$DecoderName/g, (0, generation_utils_1.createDecoderName)(definitionName))
        .replace(/\$Class/g, definitionName)
        .trim())
        .join('\n');
    const rawDecoderOutput = decodersFileTemplate
        .replace(/\$Imports/g, addFormats ? 'import addFormats from "ajv-formats"' : '')
        .replace(/\$Formats/g, addFormats ? `addFormats(ajv, ${formatOptions ? JSON.stringify(formatOptions) : 'undefined'});` : '')
        .replace(/\$ModelImports/g, definitionNames.join(', '))
        .replace(/\$Decoders/g, decoders);
    const decoderOutput = (0, prettier_1.format)(rawDecoderOutput, prettierOptions);
    outDirs.forEach(outDir => {
        (0, fs_1.mkdirSync)(outDir, { recursive: true });
        (0, fs_1.writeFileSync)(path_1.default.join(outDir, `decoders.ts`), decoderOutput);
    });
}
exports.generateCompileBasedDecoders = generateCompileBasedDecoders;
