"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generate = void 0;
// import { compile as compileTypescript } from 'json-schema-to-typescript';
const fs_1 = require("fs");
const promises_1 = require("fs/promises");
const path_1 = require("path");
const prettier_1 = require("prettier");
const controller_1 = require("../transformer/controller");
const utils_1 = require("../utils");
const controller_2 = require("./controller");
const interfaces_1 = require("./interfaces");
const schemas_1 = require("./schemas");
async function generate(options, schema) {
    const interfaces = await (0, interfaces_1.generateInterface)(options, schema.routeGroups);
    const controllers = await (0, controller_2.generateControllers)(options, schema.routeGroups);
    const schemas = await (0, schemas_1.generateSchemas)(options, schema.routeGroups);
    const groupNames = [...schema.routeGroups.keys()];
    await Promise.all(groupNames.map(async (groupName) => {
        const path = (0, path_1.resolve)(schema.version, options.outDirectory);
        const controllersPath = (0, path_1.resolve)(path, 'controllers', groupName);
        const groupUcFirst = (0, utils_1.ucFirst)(groupName);
        console.log('Creating dirs...', path);
        await (0, promises_1.mkdir)(controllersPath, { recursive: true });
        const typesPath = (0, path_1.resolve)(controllersPath, 'types.ts');
        const groupInterfaces = interfaces.get(groupName);
        let typesContent = `
                /* eslint-disable @typescript-eslint/no-empty-interface */

                /**
                 * This file is auto-generated by . Please do not modify it manually.
                  */
            `;
        const interfaceNames = [];
        groupInterfaces.forEach(groupInterface => {
            typesContent += `
                    ${groupInterface.params || ''}

                    ${groupInterface.query || ''}

                    ${groupInterface.body || ''}

                    ${groupInterface.response || ''}
                `;
            interfaceNames.push(...[
                groupInterface.names.params,
                groupInterface.names.body,
                groupInterface.names.query,
                groupInterface.names.response,
            ].filter(Boolean));
        });
        console.log('typesContent=', typesContent);
        typesContent = (0, prettier_1.format)(typesContent, options.prettierOptions);
        console.log('Writing types.ts...', typesPath, typesContent.length, 'bytes');
        await (0, promises_1.writeFile)(typesPath, typesContent, { flag: 'w' });
        const controllerPath = (0, path_1.resolve)(controllersPath, `${groupName}.controller.ts`);
        const controllerCode = controllers.get(groupName);
        controllerCode.imports.addImport('fastify-decorators', 'Controller');
        const controllerClass = `${groupUcFirst}Controller`;
        if ((0, fs_1.existsSync)(controllerPath)) {
            console.log('Controller already exists! Analysing and rewriting...');
            const oldControllerCode = await (0, promises_1.readFile)(controllerPath, 'utf-8');
            const newControllerCode = (0, prettier_1.format)((0, controller_1.transformController)(oldControllerCode, {
                className: controllerClass,
                imports: controllerCode.imports.asRecord(),
                methods: controllerCode.methods,
            }), options.prettierOptions);
            console.log('Writing controller...', controllerPath, newControllerCode.length, 'bytes');
            await (0, promises_1.writeFile)(controllerPath, newControllerCode, { flag: 'w' });
        }
        else {
            const controllerImportsText = controllerCode.imports.generate();
            const controllerContent = (0, prettier_1.format)(`
                    ${controllerImportsText}

                    @Controller({ route: '/${groupName}' })
                    export default class ${groupUcFirst}Controller {
                        ${controllerCode.inClassBody}   
                    }
                `, options.prettierOptions);
            console.log('Writing controller...', controllerPath, controllerContent.length, 'bytes');
            await (0, promises_1.writeFile)(controllerPath, controllerContent, { flag: 'w' });
        }
        const schemasPath = (0, path_1.resolve)(controllersPath, 'schemas.ts');
        const groupSchemas = schemas.get(groupName);
        const schemasContent = (0, prettier_1.format)(`
                /**
                 * This file is auto-generated by . Please do not modify it manually.
                 */
                ${groupSchemas.imports}

                ${groupSchemas.routes
            .map(e => `
                            export interface ${e.schemaTypeName} ${e.schemaTypeContent};
                            export const ${e.schemaObjectName} = ${e.schemaObjectContent};
                        `)
            .join('\n')}
            `, options.prettierOptions);
        console.log('Writing schemas.ts...', schemasPath, schemasContent.length, 'bytes');
        await (0, promises_1.writeFile)(schemasPath, schemasContent, { flag: 'w' });
    }));
}
exports.generate = generate;
