"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.transformController = void 0;
// @see https://ts-ast-viewer.com/#
const typescript_1 = __importDefault(require("typescript"));
const utils_1 = require("../utils");
function isApiDecorator(node) {
    if (typescript_1.default.isCallExpression(node.expression)) {
        if (typescript_1.default.isIdentifier(node.expression.expression)) {
            return (0, utils_1.isHTTPMethod)(node.expression.expression.escapedText.toString());
        }
    }
    return false;
}
function isApiMethod(modifiers) {
    return modifiers.some(modifier => {
        if (typescript_1.default.isDecorator(modifier)) {
            return isApiDecorator(modifier);
        }
    });
}
/**
 *
 * @param code
 * @param desiredData
 * @example const desiredData: DesiredData = {
    className: 'AuthController',
    imports: {},
    methods: {
        loginHandler: {
            reqSchemaName: 'LoginSchemaType',
            resTypeName: 'FastifyReply',
            decorators: {
                POST: {
                    url: '/login',
                    schemaName: 'LoginSchema',
                    options: {
                        bodyLimit: '1024 * 1024 * 8',
                    },
                },
            },
        },
        logoutHandler: {
            reqSchemaName: 'LogoutSchemaType',
            resTypeName: 'FastifyReply',
            decorators: {
                GET: {
                    url: '/logout',
                    schemaName: 'LogoutSchema',
                    options: {},
                },
            },
        },
    },
};
 */
function transformController(code, desiredData) {
    const sourceFile = typescript_1.default.createSourceFile('test.ts', code, typescript_1.default.ScriptTarget.Latest);
    const globalInfo = {
        methods: new Map(),
    };
    const createdMethodNames = [];
    const methodsToDelete = [];
    const classNameTransformer = (ctx) => {
        return (src) => {
            const visitor = (node) => {
                if (typescript_1.default.isClassDeclaration(node)) {
                    const nodeName = node.name;
                    if (!nodeName) {
                        throw new Error('Class must have a name');
                    }
                    if (typescript_1.default.isIdentifier(nodeName)) {
                        const className = nodeName.escapedText.toString();
                        globalInfo.className = className;
                        globalInfo.classNode = node;
                    }
                    return typescript_1.default.factory.updateClassDeclaration(node, node.modifiers, typescript_1.default.factory.createIdentifier(desiredData.className), node.typeParameters, node.heritageClauses, node.members);
                }
                return typescript_1.default.visitEachChild(node, visitor, ctx);
            };
            return typescript_1.default.visitNode(src, visitor);
        };
    };
    const createMethodsTransformer = (ctx) => {
        const desiredMethodNames = Object.keys(desiredData.methods);
        return (src) => {
            const visitor = (node) => {
                if (typescript_1.default.isClassDeclaration(node)) {
                    const { members } = node;
                    const existingMethodNames = [];
                    for (const member of members) {
                        if (typescript_1.default.isMethodDeclaration(member)) {
                            if (typescript_1.default.isIdentifier(member.name)) {
                                const name = member.name.escapedText.toString();
                                const isApi = isApiMethod(member.modifiers || []);
                                if (isApi && !desiredMethodNames.includes(name)) {
                                    // Should remove API method that is not defined
                                    methodsToDelete.push(member);
                                }
                                else {
                                    existingMethodNames.push(name);
                                }
                            }
                        }
                    }
                    const newMethods = [];
                    for (const methodName of desiredMethodNames) {
                        if (!existingMethodNames.includes(methodName)) {
                            const newBody = typescript_1.default.factory.createBlock([
                                typescript_1.default.factory.createThrowStatement(typescript_1.default.factory.createNewExpression(typescript_1.default.factory.createIdentifier('Error'), undefined, [typescript_1.default.factory.createStringLiteral('TODO: implement handler')])),
                            ], true);
                            const decorators = [];
                            const methodInfo = desiredData.methods[methodName];
                            for (const [httpMethod, decoratorData] of Object.entries(methodInfo.decorators)) {
                                const newDecorator = typescript_1.default.factory.createDecorator(typescript_1.default.factory.createCallExpression(typescript_1.default.factory.createIdentifier(httpMethod), undefined, [
                                    typescript_1.default.factory.createObjectLiteralExpression([
                                        typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('url'), typescript_1.default.factory.createStringLiteral(decoratorData.url)),
                                        typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('options'), typescript_1.default.factory.createObjectLiteralExpression([
                                            typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('schema'), typescript_1.default.factory.createIdentifier(decoratorData.schemaName)),
                                        ], true)),
                                    ], true),
                                ]));
                                decorators.push(newDecorator);
                            }
                            const newMethod = typescript_1.default.factory.createMethodDeclaration(decorators, undefined, methodName, undefined, [], [
                                typescript_1.default.factory.createParameterDeclaration(undefined, undefined, typescript_1.default.factory.createIdentifier('req'), undefined, typescript_1.default.factory.createTypeReferenceNode(typescript_1.default.factory.createIdentifier('FastifyRequest'), [
                                    typescript_1.default.factory.createTypeReferenceNode(typescript_1.default.factory.createIdentifier(methodInfo.reqSchemaName), undefined),
                                ]), undefined),
                                typescript_1.default.factory.createParameterDeclaration(undefined, undefined, typescript_1.default.factory.createIdentifier('res'), undefined, typescript_1.default.factory.createTypeReferenceNode(typescript_1.default.factory.createIdentifier(methodInfo.resTypeName), undefined), undefined),
                            ], undefined, newBody);
                            createdMethodNames.push(methodName);
                            newMethods.push(newMethod);
                        }
                    }
                    return typescript_1.default.factory.updateClassDeclaration(node, node.modifiers, node.name, node.typeParameters, node.heritageClauses, [...node.members, ...newMethods]);
                }
                return typescript_1.default.visitEachChild(node, visitor, ctx);
            };
            return typescript_1.default.visitNode(src, visitor);
        };
    };
    const patchMethodsTransformer = (ctx) => {
        return (src) => {
            const visitor = (node) => {
                if (typescript_1.default.isMethodDeclaration(node)) {
                    if (methodsToDelete.includes(node))
                        return undefined;
                    if (!typescript_1.default.isIdentifier(node.name))
                        return node;
                    const name = node.name.escapedText.toString();
                    if (createdMethodNames.includes(name))
                        return node;
                    const isApi = isApiMethod(node.modifiers || []);
                    if (!isApi)
                        return node;
                    const methodInfo = desiredData.methods[name];
                    const newModifiers = [];
                    const visitedDesiredDecorators = [];
                    for (const modifier of node.modifiers) {
                        if (typescript_1.default.isDecorator(modifier)) {
                            const isApi = isApiDecorator(modifier);
                            if (isApi) {
                                const apiCallExpr = modifier.expression;
                                if (typescript_1.default.isCallExpression(apiCallExpr)) {
                                    if (typescript_1.default.isIdentifier(apiCallExpr.expression)) {
                                        const decoratorName = apiCallExpr.expression.escapedText.toString();
                                        const decoratorInfo = methodInfo.decorators[decoratorName];
                                        visitedDesiredDecorators.push(decoratorInfo);
                                        const newDecoratorProperties = [];
                                        const arg = apiCallExpr.arguments[0];
                                        if (typescript_1.default.isObjectLiteralExpression(arg)) {
                                            for (const argProp of arg.properties) {
                                                if (typescript_1.default.isPropertyAssignment(argProp)) {
                                                    if (typescript_1.default.isIdentifier(argProp.name)) {
                                                        const fieldName = argProp.name.escapedText.toString();
                                                        if (fieldName === 'url') {
                                                            const desiredUrl = decoratorInfo?.url;
                                                            newDecoratorProperties.push(typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('url'), typescript_1.default.factory.createStringLiteral(desiredUrl)));
                                                        }
                                                        else if (fieldName === 'options') {
                                                            // TODO: combine options!!! currently leaves as it is
                                                            console.log(`Not changing ${decoratorName} options because its not easy`);
                                                            newDecoratorProperties.push(argProp);
                                                        }
                                                        else {
                                                            console.log('field', fieldName);
                                                            newDecoratorProperties.push(argProp);
                                                        }
                                                    }
                                                }
                                            }
                                        }
                                        const objLiteral = typescript_1.default.factory.createObjectLiteralExpression(newDecoratorProperties, true);
                                        newModifiers.push(typescript_1.default.factory.createDecorator(typescript_1.default.factory.createCallExpression(typescript_1.default.factory.createIdentifier(decoratorName), undefined, [objLiteral])));
                                        continue;
                                    }
                                }
                            }
                        }
                    }
                    for (const [k, v] of Object.entries(methodInfo.decorators)) {
                        if (!visitedDesiredDecorators.includes(v)) {
                            // IT IS UNVISITED! must create new!
                            const httpMethod = k;
                            const newDecorator = typescript_1.default.factory.createDecorator(typescript_1.default.factory.createCallExpression(typescript_1.default.factory.createIdentifier(httpMethod), undefined, [
                                typescript_1.default.factory.createObjectLiteralExpression([
                                    typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('url'), typescript_1.default.factory.createStringLiteral(v.url)),
                                    typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('options'), typescript_1.default.factory.createObjectLiteralExpression([
                                        typescript_1.default.factory.createPropertyAssignment(typescript_1.default.factory.createIdentifier('schema'), typescript_1.default.factory.createIdentifier(v.schemaName)),
                                    ], true)),
                                ], true),
                            ]));
                            newModifiers.push(newDecorator);
                        }
                    }
                    newModifiers.push(...node.modifiers.filter(e => !typescript_1.default.isDecorator(e)));
                    const newParameters = [
                        typescript_1.default.factory.createParameterDeclaration(undefined, undefined, typescript_1.default.factory.createIdentifier('req'), undefined, typescript_1.default.factory.createTypeReferenceNode(typescript_1.default.factory.createIdentifier('FastifyRequest'), [
                            typescript_1.default.factory.createTypeReferenceNode(typescript_1.default.factory.createIdentifier(methodInfo.reqSchemaName), undefined),
                        ]), undefined),
                        typescript_1.default.factory.createParameterDeclaration(undefined, undefined, typescript_1.default.factory.createIdentifier('res'), undefined, typescript_1.default.factory.createTypeReferenceNode(typescript_1.default.factory.createIdentifier(methodInfo.resTypeName), undefined), undefined),
                    ];
                    return typescript_1.default.factory.updateMethodDeclaration(node, newModifiers, undefined, node.name, undefined, node.typeParameters, newParameters, node.type, node.body);
                }
                return typescript_1.default.visitEachChild(node, visitor, ctx);
            };
            return typescript_1.default.visitNode(src, visitor);
        };
    };
    const { transformed } = typescript_1.default.transform(sourceFile, [classNameTransformer, createMethodsTransformer, patchMethodsTransformer], typescript_1.default.getDefaultCompilerOptions());
    const result = typescript_1.default.createPrinter({ newLine: typescript_1.default.NewLineKind.CarriageReturnLineFeed });
    return result.printFile(transformed[0]);
}
exports.transformController = transformController;
